/////////////////////////////////////////////////////////////////////////////
// mkbase.h: CComMoniker declaration
/////////////////////////////////////////////////////////////////////////////
// Copyright (c) 1997-2002, Tim Ewald, Chris Sells and Don Box.
// All rights reserved.
//
// The following is a moniker base class.
// NO WARRANTIES ARE EXTENDED. USE AT YOUR OWN RISK.
//
// To contact the authors with suggestions or comments, use
// tewald@obelisk-llc.com,
// csells@sellsbrothers.com and
// dbox@develop.com.
// Usage:
/*
class ATL_NO_VTABLE <%= className %> : 
	public CComObjectRootEx<CComSingleThreadModel>,
	public CComCoClass<<%= className %>, &<%= clsid %>>,
    public CComMoniker<<%= className %>>
{
public:
DECLARE_REGISTRY_RESOURCEID(<%= resourceId %>)
DECLARE_PROTECT_FINAL_CONSTRUCT()

BEGIN_COM_MAP(<%= className %>)
	COM_INTERFACE_ENTRY(IPersist)
	COM_INTERFACE_ENTRY(IPersistStream)
	COM_INTERFACE_ENTRY(IMoniker)
	COM_INTERFACE_ENTRY(IParseDisplayName)
	COM_INTERFACE_ENTRY(IROTData)
	COM_INTERFACE_ENTRY(IMarshal)
END_COM_MAP()

// CComMoniker overrides
public:
    STDMETHODIMP BindToObject(IBindCtx *pbc, IMoniker *pmkToLeft, REFIID riidResult, void **ppvResult);
	STDMETHODIMP GetDisplayName(IBindCtx *pbc, IMoniker *pmkToLeft, LPOLESTR *ppszDisplayName);
	STDMETHODIMP ParseDisplayName(IBindCtx *pbc, IMoniker *pmkToLeft, LPOLESTR pszDisplayName, ULONG *pchEaten, IMoniker **ppmkOut);
};
*/
/////////////////////////////////////////////////////////////////////////////
// History:
// 2/9/01:  Moved implementation to header for convenience.
// 3/7/00:  Simplified usage.
// 6/26/97: 1st release.

#ifndef __COMMONIKER_H__
#define __COMMONIKER_H__

#include <windows.h>

#ifndef lengthof
#define lengthof(_rg) (sizeof(_rg)/sizeof(*_rg))
#endif

/////////////////////////////////////////////////////////////////////////////
// CComMonikerBase

// CComMonikerRootBase implements the reduction
// and composition methods in the "default" way,
// that is, it assumes you do nothing special
// in terms of composition or reduction

class CComMonikerRootBase : public IMoniker
{
public:
    // IMoniker
	STDMETHODIMP Reduce(IBindCtx *pbc, DWORD dwReduceHowFar,
                        IMoniker **ppmkToLeft, IMoniker **ppmkReduced);
	STDMETHODIMP ComposeWith(IMoniker *pmkRight, BOOL fOnlyIfNotGeneric,
                             IMoniker **ppmkComposite);
    STDMETHODIMP Inverse(IMoniker **ppmk);
	STDMETHODIMP CommonPrefixWith(IMoniker *pmkOther, IMoniker **ppmkPrefix);
	STDMETHODIMP RelativePathTo(IMoniker *pmkOther, IMoniker **ppmkRelPath);
};

class CComMonikerBase :
	public CComMonikerRootBase,
    public IParseDisplayName,
	public IROTData,
	public IMarshal
{
public:
    CComMonikerBase();
    virtual ~CComMonikerBase();

    // IUnknown (so we don't have to depend on ATL)
    STDMETHODIMP QueryInterface(REFIID riid, void** ppv) =0;

    // IPersist (implemented in CComMoniker)
	STDMETHODIMP GetClassID(CLSID *pClassID) =0;

    // IPersistStream
	STDMETHODIMP IsDirty();
    STDMETHODIMP Load(IStream *pStm);
	STDMETHODIMP Save(IStream *pStm, BOOL fClearDirty);
	STDMETHODIMP GetSizeMax(ULARGE_INTEGER *pcbSize);

    // IMarshal
	STDMETHODIMP GetUnmarshalClass(REFIID riid, void *pv, DWORD dwDestContext,
                                   void *pvDestContext, DWORD mshlflags, CLSID *pCid);
	STDMETHODIMP GetMarshalSizeMax(REFIID riid, void *pv, DWORD dwDestContext,
                                   void *pvDestContext, DWORD mshlflags, DWORD *pSize);
	STDMETHODIMP MarshalInterface(IStream *pStm, REFIID riid, void *pv, DWORD dwDestContext,
                                  void *pvDestContext, DWORD mshlflags);

	STDMETHODIMP UnmarshalInterface(IStream *pStm, REFIID riid, void **ppv);
	STDMETHODIMP ReleaseMarshalData(IStream *pStm);
	STDMETHODIMP DisconnectObject(DWORD dwReserved);

    // IParseDisplayName    
    STDMETHODIMP ParseDisplayName(IBindCtx *pbc, LPOLESTR pszDisplayName,
                                  ULONG *pchEaten, IMoniker **ppmkOut);

    // IROTData
	STDMETHODIMP GetComparisonData(byte* pbData, ULONG cbMax, ULONG *pcbData);

    // IMoniker
	STDMETHODIMP BindToStorage(IBindCtx *pbc, IMoniker *pmkToLeft,
                               REFIID riid, void **ppvObj);

	STDMETHODIMP Enum(BOOL fForward, IEnumMoniker **ppenumMoniker);
	STDMETHODIMP IsEqual(IMoniker *pmkOtherMoniker);
	STDMETHODIMP Hash(DWORD *pdwHash);

	STDMETHODIMP IsRunning(IBindCtx *pbc, IMoniker *pmkToLeft, IMoniker *pmkNewlyRunning);
	STDMETHODIMP GetTimeOfLastChange(IBindCtx *pbc, IMoniker *pmkToLeft, FILETIME *pFileTime);
	STDMETHODIMP IsSystemMoniker(DWORD *pdwMksys);

    // YOU MUST IMPLEMENT!
    STDMETHODIMP BindToObject(IBindCtx *pbc, IMoniker *pmkToLeft,
                              REFIID riidResult, void **ppvResult) =0;

	STDMETHODIMP GetDisplayName(IBindCtx *pbc, IMoniker *pmkToLeft,
                                LPOLESTR *ppszDisplayName) =0;

	STDMETHODIMP ParseDisplayName(IBindCtx *pbc, IMoniker *pmkToLeft,
                                  LPOLESTR pszDisplayName, ULONG *pchEaten,
                                  IMoniker **ppmkOut) =0;

protected:
    // Display Name Helpers
    virtual const wchar_t* ProgID() =0;
    virtual const wchar_t* VersionIndependentProgID() =0;
    
    // Match "ProgID:" or "VersionIndependentProgID:"
    // and return pointer one character past ":" in *ppwszDisplayParam
    bool MatchesProgID(const wchar_t* pwszDisplayName, const wchar_t** ppwszDisplayParam);

private:
    const OLECHAR*  m_pwszDisplayName;

    HRESULT CacheDisplayName();
};

template <typename Deriving>
class CComMoniker : public CComMonikerBase
{
public:
    CComMoniker() : m_pszProgID(0) { *m_szVIProgID = 0; }
    virtual ~CComMoniker() { CoTaskMemFree(m_pszProgID); }

	STDMETHODIMP GetClassID(CLSID *pClassID)
    {
        if( !pClassID ) return E_POINTER;
        *pClassID = This()->GetObjectCLSID();
        return S_OK;
    }

protected:
    // Display Name Helpers
    // NOTE: ProgID has a version, e.g. foo.bar.1,
    //       whereas VIProgID does not, e.g. foo.bar,
    //       and ProgIDFromCLSID returns the version'd progid

    virtual const wchar_t* ProgID()
    {
        if( !m_pszProgID ) ProgIDFromCLSID(This()->GetObjectCLSID(), &m_pszProgID);
        return m_pszProgID;
    }

    virtual const wchar_t* VersionIndependentProgID()
    {
        if( !*m_szVIProgID )
        {
            TCHAR   szKey[128];
            wsprintf(szKey, __T("CLSID\\%S\\VersionIndependentProgID"), static_cast<BSTR>(CComBSTR(This()->GetObjectCLSID())));

            TCHAR   szVIProgID[lengthof(m_szVIProgID)+1];
            long    cbVIProgID = sizeof(szVIProgID);
            long    ret = ::RegQueryValue(HKEY_CLASSES_ROOT, szKey, szVIProgID, &cbVIProgID);
            if( (ret == ERROR_SUCCESS) && cbVIProgID )
            {
                // TODO: Won't work under Unicode build
                mbstowcs(m_szVIProgID, szVIProgID, lengthof(m_szVIProgID)+1);
            }
            else if( ProgID() )
            {
                wcscpy(m_szVIProgID, m_pszProgID);
            }
        }

        return m_szVIProgID;
    }

    Deriving* This() { return static_cast<Deriving*>(this); }

private:
    OLECHAR     m_szVIProgID[256];  // No version, e.g. foo.bar
    LPOLESTR    m_pszProgID;        // Version, e.g. foo.bar.1
};

// Inline Implementations
inline
CComMonikerBase::CComMonikerBase()
    : m_pwszDisplayName(0)
{}

inline
CComMonikerBase::~CComMonikerBase()
{ CoTaskMemFree(const_cast<OLECHAR*>(m_pwszDisplayName)); }

inline
STDMETHODIMP CComMonikerRootBase::Reduce(
    IBindCtx*   pbc,
    DWORD       dwReduceHowFar,
    IMoniker**  ppmkToLeft,
    IMoniker**  ppmkReduced)
{
    HRESULT hr = E_POINTER;
    if( ppmkReduced )
    {
        if( ppmkToLeft ) *ppmkToLeft = 0;
        (*ppmkReduced = this)->AddRef();
        hr = MK_S_REDUCED_TO_SELF;
    }

    return hr;
}

inline
STDMETHODIMP CComMonikerRootBase::ComposeWith(
    IMoniker*   pmkRight,
    BOOL        fOnlyIfNotGeneric,
    IMoniker**  ppmkComposite)
{
    HRESULT hr = E_INVALIDARG;
    if( ppmkComposite )
    {
        *ppmkComposite = 0;
        if( pmkRight )
        {
            DWORD dwMkSys;
            hr = pmkRight->IsSystemMoniker(&dwMkSys);
            if( !(hr == S_OK && dwMkSys == MKSYS_ANTIMONIKER) )
            {
                if( fOnlyIfNotGeneric )
                    hr = MK_E_NEEDGENERIC;
                else
                    hr = CreateGenericComposite(this, pmkRight, ppmkComposite);
            }
            else
                hr = S_OK; // right is AntiMoniker, so we're done
        }
    }
    return hr;
}

inline
STDMETHODIMP CComMonikerRootBase::Inverse(
    IMoniker**  ppmk)
{
    return CreateAntiMoniker(ppmk);
}

inline
STDMETHODIMP CComMonikerRootBase::CommonPrefixWith(
    IMoniker*   pmkOther,
    IMoniker**  ppmkPrefix)
{
    HRESULT hr = E_INVALIDARG;
    if( ppmkPrefix )
    {
        *ppmkPrefix = 0;
        if( pmkOther )
        {
            hr = this->IsEqual(pmkOther);
            if( hr == S_OK )
            {
                (*ppmkPrefix = this)->AddRef();
                hr = MK_S_US;
            }
            else
                hr = MonikerCommonPrefixWith(this, pmkOther, ppmkPrefix);
        }
    }
    return hr;
}

inline
STDMETHODIMP CComMonikerRootBase::RelativePathTo(
    IMoniker*   pmkOther, 
    IMoniker**  ppmkRelPath)
{
    HRESULT hr = E_INVALIDARG;
    if( ppmkRelPath != 0 )
    {
        *ppmkRelPath = 0;
        hr = MonikerRelativePathTo(this, pmkOther, ppmkRelPath, TRUE);
    }
    return hr;
}

inline
HRESULT CComMonikerBase::CacheDisplayName()
{
    HRESULT hr = S_OK;

    // If we don't have the constant display named
    // cached, ask the derived class for it.
    if( !m_pwszDisplayName )
    {
        CComPtr<IBindCtx>   pbc;
        hr = CreateBindCtx(0, &pbc);

        if( SUCCEEDED(hr) )
        {
            // Const-cast is required as Monikers are
            // set once and then read-only.
            hr = GetDisplayName(pbc, 0,
                                const_cast<OLECHAR**>(&m_pwszDisplayName));
        }
    }

    return hr;
}

// IPersistStream

inline
STDMETHODIMP CComMonikerBase::IsDirty()
{
    // Monikers are constant (hence the by-value marshaling)
    return S_FALSE;
}

inline
STDMETHODIMP CComMonikerBase::Load(
    IStream* pstm)
{
    HRESULT hr;

    if( !pstm ) return E_INVALIDARG;
    if( m_pwszDisplayName ) return E_UNEXPECTED;

    // TODO: Support other platforms
    DWORD   cb;
    DWORD   nTag;
    hr = pstm->Read(&nTag, sizeof(nTag), &cb);
    if( FAILED(hr) ) return hr;
    if( nTag != NDR_LOCAL_DATA_REPRESENTATION ) return E_UNEXPECTED;

    DWORD   nSize;
    hr = pstm->Read(&nSize, sizeof(nSize), &cb);
    if( FAILED(hr) ) return hr;

    // Const-cast is required as Monikers are
    // set once and then read-only.
    m_pwszDisplayName = (OLECHAR*)CoTaskMemAlloc(nSize);
    if( !m_pwszDisplayName ) return E_OUTOFMEMORY;

    hr = pstm->Read(const_cast<OLECHAR*>(m_pwszDisplayName), nSize, &cb);
    if( FAILED(hr) ) return hr;

    ULONG               cch;
    CComPtr<IBindCtx>   pbc;
    CComPtr<IMoniker>   pmk;

    hr = CreateBindCtx(0, &pbc);
    if( FAILED(hr) ) return hr;

    // const_cast<> required because
    // ParseDisplayName() isn't const correct!
    hr = ParseDisplayName(pbc, 0,
                          const_cast<OLECHAR*>(m_pwszDisplayName),
                          &cch, &pmk);

    return hr;
}

inline
STDMETHODIMP CComMonikerBase::Save(
    IStream*    pstm,
    BOOL        fClearDirty)
{
    HRESULT hr;
    if( !pstm ) return E_INVALIDARG;

    hr = CacheDisplayName();
    if( SUCCEEDED(hr) )
    {
        DWORD   nSize = (wcslen(m_pwszDisplayName) + 1) * sizeof(OLECHAR);
        DWORD   nTag = NDR_LOCAL_DATA_REPRESENTATION;
        DWORD   cb;

        if( SUCCEEDED(hr = pstm->Write(&nTag, sizeof(nTag), &cb)) &&
            SUCCEEDED(hr = pstm->Write(&nSize, sizeof(nSize), &cb)) )
        {
            hr = pstm->Write(m_pwszDisplayName, nSize, &cb);
        }
    }

    return hr;
}

inline
STDMETHODIMP CComMonikerBase::GetSizeMax(
    ULARGE_INTEGER* pcbSize)
{
    HRESULT hr;
    if( !pcbSize ) return E_POINTER;

    hr = CacheDisplayName();
    if( SUCCEEDED(hr) )
    {
        pcbSize->LowPart = sizeof(DWORD) +  // NDR data rep.
                           sizeof(DWORD) +  // Display name size
                           wcslen(m_pwszDisplayName) * sizeof(OLECHAR) +  // Display name
                           sizeof(OLECHAR); // NULL
        pcbSize->HighPart = 0;
    }

    return hr;
}

// IMarshal

inline
STDMETHODIMP CComMonikerBase::GetUnmarshalClass(
    REFIID  riid,
    void*   pv,
    DWORD   dwDestContext,
    void*   pvDestContext,
    DWORD   mshlflags,
    CLSID*  pCid)
{
    return GetClassID(pCid);
}

inline
STDMETHODIMP CComMonikerBase::GetMarshalSizeMax(
    REFIID  riid,
    void*   pv,
    DWORD   dwDestContext,
    void*   pvDestContext,
    DWORD   mshlflags,
    DWORD*  pSize)
{
    HRESULT hr;

    if( !pSize ) return E_POINTER;
    *pSize = 0;

    ULARGE_INTEGER  uli;
    hr = GetSizeMax(&uli);
    if( SUCCEEDED(hr) )
    {
        *pSize = uli.LowPart;
    }

    return hr;
}

inline
STDMETHODIMP CComMonikerBase::MarshalInterface(
    IStream*    pstm,
    REFIID      riid,
    void*       pv,
    DWORD       dwDestContext,
    void*       pvDestContext,
    DWORD       mshlflags)
{
    return Save(pstm, FALSE);
}

inline
STDMETHODIMP CComMonikerBase::UnmarshalInterface(
    IStream*    pstm,
    REFIID      riid,
    void**      ppv)
{
    HRESULT hr;
    if( !ppv ) return E_POINTER; *ppv = 0;

    hr = Load(pstm);
    if( SUCCEEDED(hr) )
    {
        hr = QueryInterface(riid, ppv);
    }

    return hr;
}

inline
STDMETHODIMP CComMonikerBase::ReleaseMarshalData(
    IStream*    /*pstm*/)
{
    return S_OK;
}

inline
STDMETHODIMP CComMonikerBase::DisconnectObject(
    DWORD   /*dwReserved*/)
{
    return S_OK;
}

// IMoniker

inline
STDMETHODIMP CComMonikerBase::BindToStorage(
    IBindCtx*   /*pbc*/,
    IMoniker*   /*pmkToLeft*/,
    REFIID      /*riid*/,
    void**      ppvObj)
{
    if( !ppvObj ) return E_POINTER;
    *ppvObj = 0;
    return E_NOTIMPL;
}

inline
STDMETHODIMP CComMonikerBase::Enum(
    BOOL            fForward,
    IEnumMoniker**  ppenumMoniker)
{
    if( !ppenumMoniker ) return E_POINTER;
    *ppenumMoniker = 0;
    return S_OK;
}

inline
STDMETHODIMP CComMonikerBase::IsEqual(
    IMoniker*   pmkOtherMoniker)
{
    HRESULT hr;

    // Check the type of the other moniker for us
    CLSID   clsid;
    CLSID   clsidOther;

    if( SUCCEEDED(hr = GetClassID(&clsid)) &&
        SUCCEEDED(hr = pmkOtherMoniker->GetClassID(&clsidOther)) )
    {
        if( clsid == clsidOther )
        {
            CComQIPtr<IROTData> prd = pmkOtherMoniker;
            if( prd )
            {
                byte	rgb[2][2048];
                ULONG	cbData[2];
    
                if( SUCCEEDED(hr = GetComparisonData(rgb[0], 2048, cbData + 0)) &&
                    SUCCEEDED(hr = prd->GetComparisonData(rgb[1], 2048, cbData + 1)) )
                {
                    if( cbData[0] == cbData[1] &&
                        memcmp(rgb[0], rgb[1], cbData[0]) == 0 )
                    {
                        hr = S_OK;
                    }
                    else
                    {
                        hr = S_FALSE;
                    }
                }
            }
        }
        else
        {
            hr = S_FALSE;
        }
    }

    return hr;
}

inline
STDMETHODIMP CComMonikerBase::Hash(
    DWORD*  pdwHash)
{
    HRESULT hr = S_OK;
    if( !pdwHash ) return E_POINTER; *pdwHash = 0;
    if( !m_pwszDisplayName ) return E_UNEXPECTED;

    // TODO: Pick something better than a sum
    const OLECHAR*  pwsz = m_pwszDisplayName;
    while( *pwsz )
    {
        *pdwHash += *pwsz;
        pwsz++;
    }

    return hr;
}

inline
STDMETHODIMP CComMonikerBase::IsRunning(
    IBindCtx*   /*pbc*/,
    IMoniker*   /*pmkToLeft*/,
    IMoniker*   /*pmkNewlyRunning*/)
{
    return S_FALSE;
}

inline
STDMETHODIMP CComMonikerBase::GetTimeOfLastChange(
    IBindCtx*   /*pbc*/,
    IMoniker*   /*pmkToLeft*/,
    FILETIME*   /*pFileTime*/)
{
    return MK_E_UNAVAILABLE;
}

inline
STDMETHODIMP CComMonikerBase::ParseDisplayName(
    IBindCtx*   pbc,
    LPOLESTR    pszDisplayName,
    ULONG*      pchEaten,
    IMoniker**  ppmkOut)
{
    return ParseDisplayName(pbc, 0, pszDisplayName, pchEaten, ppmkOut);
}

inline
STDMETHODIMP CComMonikerBase::IsSystemMoniker(
    DWORD*  pdwMksys)
{
    *pdwMksys = MKSYS_NONE;
    return S_FALSE;
}

// IROTData

inline
STDMETHODIMP CComMonikerBase::GetComparisonData(
    byte*   pbData,
    ULONG   cbMax,
    ULONG*  pcbData)
{
    HRESULT hr = S_OK;
    if( !pbData ) return E_POINTER; *pbData = 0;
    if( !pcbData ) return E_POINTER; *pcbData = 0;
    if( !m_pwszDisplayName ) return E_UNEXPECTED;

    ULONG&  cbData = *pcbData;
    cbData = wcslen(m_pwszDisplayName);
    if( cbData > cbMax ) cbData = cbMax; 
    memcpy(pbData, m_pwszDisplayName, cbData);

    return hr;
}

inline
bool CComMonikerBase::MatchesProgID(
    const wchar_t*  pwszDisplayName,
    const wchar_t** ppwszDisplayParam)
{
    int nColon = 0;

    if( (nColon = wcslen(ProgID()))&&
        (wcsnicmp(pwszDisplayName, ProgID(), nColon) == 0) &&
        (pwszDisplayName[nColon] == ':') )
    {
        // Fall through
    }
    else if( (nColon = wcslen(VersionIndependentProgID())) &&
             (wcsnicmp(pwszDisplayName, VersionIndependentProgID(), nColon) == 0) &&
             (pwszDisplayName[nColon] == ':') )
    {
        // Fall through
    }
    else
    {
        return false;
    }

    if( ppwszDisplayParam )
    {
        *ppwszDisplayParam = pwszDisplayName + nColon + 1;
    }

    return true;
}

#endif //__COMMONIKER_H__
